# Copyright (c) 2001-2007, RedIRIS. All Rights Reserved.
#
# You may distribute under the terms of the GNU General Public License,
# as specified in the LICENSE file that was shipped with this distribution.

package PAPI::Main;

use PAPI::ApachePoA;
use PAPI::Cookie_Handler;

use MIME::Base64 qw/decode_base64 encode_base64/;
use Crypt::OpenSSL::RSA;
use Crypt::CBC;
use Crypt::Rijndael;
use URI::Escape;
use Data::Dumper;
use Apache2::Module ();
use Apache2::Const;
use Apache2::RequestUtil ();
use Apache2::RequestRec ();

use Apache2::Request;

sub handler 
{
	my $r = shift;
	return DECLINED unless $r->is_initial_req;
	my $url=$r->uri;
	my $qst=$r->args;
	my %fields=();
	for my $par (split /&/,$qst) {
		my ($pn,$pv)=split /=/,$par;
		$fields{$pn}=$pv;
	}
	my $hli=$fields{'PAPIHLI'};
	my $opo=$fields{'PAPIOPOA'};

#	$r->no_cache(1);
	
	my $now=time;


	my $poa=new PAPI::ApachePoA;
 	### CONFIGURE PoA #############################
	$poa->set_parameters($r);
	$poa->dumpApachePoA("PoADump") if ($poa->{'Debug'});

	if ( !defined($poa->{Serv}) ){
	   	$poa->debug("PoA could not be instantiated. Aborting...");
	   	return SERVER_ERROR;
	}
	$poa->debug("Parameters initialized:" .  Dumper($poa)) if ($poa->{'Debug'});
	my $curl = $url;
	$curl .= "?". $qst if (defined $qst && $qst ne "");
	$poa->debug("Processing request: $curl") if ($poa->{'Debug'});
	# Check whether the request matches passthrough filter(s)
	$poa->debug("Checking the Pass_URL_Patterns") if ($poa->{'Debug'});
	for my $poapt (@{$poa->{PassUrlPatterns}}) { 
		if ($curl =~ /$poapt/) {
			$poa->warning("Request for $curl matches passtrhough filter $poapt. Access granted");
			# Register proxy handler if PoA is in proxy mode
			if (defined $poa->{'Remote_Serv'} || defined $poa->{'Remote_Dom'}) {
				$r->handler("perl-script");
				$r->set_handlers(PerlResponseHandler => [PAPI::RewritingProxy]);
				if (defined $poa->{'Remote_Serv'}) { $poa->accesslog("Resolving request $url to $poa->{'Remote_Serv'}$poa->{'Location'}"); }
				else { $poa->accesslog("Resolving request $url to domain $poa->{'Remote_Dom'}"); }
			}
			return DECLINED;
		}	 
	}
	
	if ($r->method() eq 'POST'){
		$poa->debug("Checking the Pass_POST_Patterns") if ($poa->{'Debug'});
		for my $poapt (@{$poa->{PassPostPatterns}}) {
    		my $req = Apache2::Request->new($r);
    		my $table = $req->body();
    		$poa->debug("Apache2:Request TABLE: <".Dumper($table).">") if ($poa->{'Debug'});
    		my $buf = '';
    		foreach $key (keys %{$table}){
    			$buf .= $table->{$key};
    		}
			$poa->debug("Comparing POST content <".$buf.">") if ($poa->{'Debug'});
    		$poa->debug("Comparing POAPT <$poapt>") if ($poa->{'Debug'});
			if ($buf =~ /$poapt/) {
        		$poa->warning("Request for $curl matches passtrhough POST filter $poapt. Access granted");
        		# Register proxy handler if PoA is in proxy mode
        		if (defined $poa->{'Remote_Serv'} || defined $poa->{'Remote_Dom'}) {
          			$r->handler("perl-script");
          			$r->set_handlers(PerlResponseHandler => [PAPI::RewritingProxy]);
          			if (defined $poa->{'Remote_Serv'}) { $poa->accesslog("Resolving request $url to $poa->{'Remote_Serv'}$poa->{'Location'}"); }
          			else { $poa->accesslog("Resolving request $url to domain $poa->{'Remote_Dom'}"); }
        		}
        		return DECLINED;
      		}
		}
	}

	# This require must come here to look for Athens:DA only in 
	# case we are using PADATH
	if (defined $poa->{Athens_Key}) { require PAPI::PADATH; }
	if ($url =~ /$poa->{Auth_Location}/){
# Redirection to an AS to reauthenticate user
		if ($url =~ /PAPIASRedirector/) {
			my $as=uri_unescape($fields{'ASID'});
			if (!defined $as || !defined $poa->{ases}{$as}) {
				return FORBIDDEN;
			}
			my $requested_url=uri_unescape($fields{'REQURL'});
			if (defined $requested_url) {
				$poa->warning("User redirected to the AS ".$as." for accessing the url ".$requested_url);
			}
			my $asurl=$poa->{ases}{$as}->{url}."?ATTREQ=".$poa->{'Serv'}."&PAPIPOAREF=".uri_escape($fields{'PAPIPOAREF'},'\W')."&PAPIPOAURL=".$fields{'PAPIPOAURL'};
			$asurl.="&PAPIHLI=".$hli if (defined $hli);
			$asurl.="&PAPIOPOA=".$opo if (defined $opo);
			$poa->debug("Redirecting user back to AS for further validation: $asurl") if ($poa->{'Debug'});
			PAPI::Cookie_Handler::redirect_URL($r,$asurl);
			return DONE;
		}
# A request for cookies coming from an AS
		else {
			$r->handler("perl-script");
			$r->set_handlers(PerlAccessHandler => undef);
			$r->set_handlers(PerlResponseHandler => [PAPI::Cookie_Handler]);
			return DECLINED;
		}
	}
	my $ohcook="0:NODATA:0:randomBlock:.:serviceID@ASID" ;
	my $code;

	### PROCESSING ACTIONS ######################################	

	$poa->{'log'}->debug("PAPI#Debug: Main Processing request: $url") if ($poa->{'Debug'});
	

	if ($fields{'ACTION'} eq "CHECKED"){

	### Answer from a GPoA ####

		my ($req_id,$valid_date); 
		### Check from an answer of an AS to an ATTREQ ###
		my $isas=$fields{'AS'};
		($req_id,$code,$valid_date)=$poa->test_GPoA_URL($fields{'DATA'},$isas);
		my $data_req=$poa->load_request($req_id);
# If no request found, assume this is a visit to a previously stored URL
# (reloaded by the user or in some history list at the browser)
		if (!defined $data_req) {
			my $ras = '?';
			for my $ia (keys %fields) {
				if ($ia ne "ACTION" && $ia ne "DATA") {
					$ras.=$ia.'='.$fields{$ia}.'&';
				}
			}
			$url .= $ras;
			chop $url;
			PAPI::Cookie_Handler::redirect_URL($r, $url);
			return DECLINED;
		}
		if ($valid_date == 0){
			$poa->warning("Invalid GPoA/AS Answer for $url");
			return FORBIDDEN;
		}
		my $Hcook=undef;
		if ($poa->{'useHcook'}) {
			$Hcook=$poa->gen_Hcook($code,$valid_date,1);
			$ohcook=$poa->{'rawHcook'};
			$poa->debug("Main: New Hcook: $Hcook") if ($poa->{'Debug'});
		}
		my $Lcook=$poa->gen_Lcook($code);
		$poa->debug("Main: New Lcook: $Lcook") if ($poa->{'Debug'});
# New cookie handling makes this unnecessary
#		$r->header_in('If-modified-since' => undef);
#		$r->header_in('If-none-match' => undef);
		my $err_out=$r->err_headers_out;
		if (defined $Hcook) { $err_out->add("Set-Cookie" => "$Hcook"); }
		$err_out->add("Set-Cookie" => "$Lcook");

		my %headers=%{ $data_req->{headers} };
		foreach my $k (keys %headers){
			if ($k eq 'Cookie') {
				my $fcv = "";
				for my $ckv (split /[\s]*;[\s]*/,$headers{$k}) {
					my ($cn, $cv) = split /[\s]*=[\s]*/,$ckv,2;
					if ($cn !~ /Hcook/i && $cn !~ /Lcook/i) {
						if ($fcv ne "") { $fcv .= "; $cn=$cv"; }
						else { $fcv = "$cn=$cv"; }
					}
				}
				$r->headers_in->set('Cookie' => $fcv) if ($fcv ne "");
			}
# New cookie handling makes this unnecessary
#			elsif ($k !~ /If-modified-since/i && $k !~ /If-none-match/i){
				$r->headers_in->set($k =>$headers{$k});
#			}
		}
		$r->method($data_req->{method});
		$r->uri($data_req->{uri});
		$r->args($data_req->{args});
		$r->filename($data_req->{filename});
		if ($data_req->{method} eq 'POST'){
			$poa->debug("Main: POST request") if ($poa->{'Debug'});
			my ($Lcook_val) = $Lcook =~ /$poa->{Lcook_Name}=([^;]+)/;
			$r->headers_in->set("Cookie" => "$poa->{Lcook_Name}=" . $Lcook_val . ";" . $r->headers_in->get('Cookie'));
			$r->pnotes('CONTENT_req',$data_req->{content});			
			$r->handler("perl-script");
			my @hdls = @{$r->get_handlers('PerlResponseHandler')};
			unshift @hdls, PAPI::POST_Handler;
			$r->set_handlers(PerlResponseHandler => \@hdls);
		}
	}
	else{
		### NORMAL REQUEST ###
		$poa->debug("Processing request: $url") if ($poa->{'Debug'});
		my ($Hcook,$Lcook)=$poa->get_cookies();
		$ohcook = $Hcook if (defined $Hcook);
		if ($poa->{'Debug'}) {
			$poa->debug("Main: Hcook received:\n #$Hcook#") if (defined $Hcook);
			$poa->debug("Main: Lcook received:\n #$Lcook#");
		}
		my $ret;
		($ret,$code)=$poa->test_Lcook($Lcook, $hli);
		if($ret==0){
			if (!defined $Hcook) {
				if ($poa->{GPoA_URL} ne "") {
					$r->handler("perl-script");
					$r->set_handlers(PerlAccessHandler => undef);
					$r->set_handlers(PerlResponseHandler => [PAPI::RedirectGPoA]);
					return DECLINED;
					}
				else {
					$poa->warning("Forbidden access to $url");
					return FORBIDDEN;	
				}
			}
			else {
				my ($valid_date,$temp,$ncrc,$gvd);
				($code,$valid_date,$temp,$ncrc,$gvd)=$poa->test_Hcook($Hcook,1, $hli);
				if($valid_date<=0){
					if ($valid_date==0 && $poa->{GPoA_URL} ne ""){
						$r->handler("perl-script");
						$r->set_handlers(PerlAccessHandler => undef);
						$r->set_handlers(PerlResponseHandler => [PAPI::RedirectGPoA]);
						return DECLINED;
					}
					else {
						$poa->warning("Forbidden access to $url");
						return FORBIDDEN;
					}
				}
				if($valid_date != 1){
					$poa->debug("Main: Generating HcooK=$code,$valid_date,$temp,$ncrc") if ($poa->{'Debug'});
					$Hcook=$poa->gen_Hcook($code,$valid_date,$temp,$ncrc);
					$poa->debug("Main: New Hcook: $Hcook") if ($poa->{'Debug'});
					$Lcook=$poa->gen_Lcook($code);
					$poa->debug("Main: New Lcook: $Lcook") if ($poa->{'Debug'});
# New cookie handling makes this unnecessary
#					$r->header_in('If-modified-since' => undef);
#					$r->header_in('If-none-match' => undef);
					my $err_out=$r->err_headers_out;
					$err_out->add("Set-Cookie" => "$Hcook");
					$err_out->add("Set-Cookie" => "$Lcook");
				}
			}
		}
		else { 
			$poa->debug("Lcook OK: $code") if ($poa->{'Debug'}); 
		}
	}
	$poa->accesslog("Access OK to $poa->{Serv}: $code: $url");
# Signoff request
	for my $sop (@{$poa->{signoffLocs}}) {
		if ($curl =~ /$sop->{regex}/) {
			my $rurl = $sop->{url};
			$poa->warning("Signoff requested through $curl. Tokens invalidated and request redirected to $rurl");
			my ($ihc, $ilc) = $poa->gen_logout_cookies($code);
			PAPI::Cookie_Handler::redirect_with_cookies ($r, $rurl, $ihc, $ilc);
			return DECLINED;
		}
	}	
# Pass user attribute information
	if ($#{$poa->{'attrList'}} == -1) { $poa->build_attrList ($code); }
	my @atrl=@{$poa->{'attrList'}};
	for (my $i=0; $i<$#atrl; $i+=2) {
		$r->pnotes("PAPIAttr-".$atrl[$i] => $atrl[$i+1]);
		$r->headers_in->set("X-PAPIAttr-".$atrl[$i] => $atrl[$i+1]);
	}		
# Pass complete token data to Hcook_Handler (if defined)
	if (defined $poa->{Hcook_Handler} && $url =~ /$poa->{Hcook_Handler}/) {
		my $nval = $ohcook;
		if ($poa->{useHcook}) {
			my $enc_packet=decode_base64($ohcook);
    		my $iv = substr($enc_packet,0,16);
    		$enc_packet= substr($enc_packet,16);
     		my $key_aes=$poa->{'Hkey'};
   			my $dec_obj = new Crypt::CBC({ key=>$key_aes,
					   cipher=>'Rijndael',
					   iv=>$iv,
       		 		   regenerate_key=>0,
					   padding=>'null',
					   prepend_iv=>0
					   });
			$nval=$dec_obj->decrypt($enc_packet); 
		}
		$r->pnotes('PAPIHcook' => $nval);
		$r->headers_in->set('X-PAPI-Hcook' => $nval);
		$poa->debug("Main: Passing Hcook value $nval to handler $url") if ($poa->{Debug});
	}
# AthensDA response in LAA mode
	if (defined $poa->{PADATH_LAA} && $url =~ /PADATHLAARespHandler/) {
		$poa->debug("Main: Handling response from LAA request by means of PADATH") if ($poa->{Debug});
		return PAPI::PADATH::respLAA($r,$poa);
	}
# A request for a URL acting as the trigger for PADATH in LAA mode
	elsif (defined $poa->{PADATH_LAA} && $url =~ /$poa->{PADATH_LAA}/) {
		$poa->debug("Main: Handling request for LAA-enabling URL by means of PADATH") if ($poa->{Debug});
		return PAPI::PADATH::reqLAA($r,$poa);
	}
# A request for a URL acting as a responder for PADATH in HDD mode
	elsif (defined $poa->{PADATH_HDD} && $url =~ /$poa->{PADATH_HDD}/) {
		$poa->debug("Main: Building response to HDD request URL by means of PADATH") if ($poa->{Debug});
		return PAPI::PADATH::respHDD($r,$poa);
	}

	my @hdls = @{$r->get_handlers('PerlAccessHandler')};
	if (!$#hdls) { $r->set_handlers(PerlAccessHandler => undef); }
	else {
		my @nhdls = ();
		for my $hd (@hdls) { push @nhdls, $hd if ($hd != \PAPI::Main); }
		$r->set_handlers(PerlAccessHandler => \@nhdls);
	}

	if (defined $poa->{'Remote_Serv'} || defined $poa->{'Remote_Dom'}) {
# Register proxy handler
		$r->handler("perl-script");
		$r->set_handlers(PerlResponseHandler => [PAPI::RewritingProxy]);
		if (defined $poa->{'Remote_Serv'}) { $poa->accesslog("Resolving request $url to $poa->{'Remote_Serv'}$poa->{'Location'}"); }
		else { $poa->accesslog("Resolving request $url to domain $poa->{'Remote_Dom'}"); }
		return DECLINED;
	}
	else { return DECLINED; }
}

1;

__END__
